/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2008 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;
import org.objectweb.fractal.bf.AbstractBindingFactoryPlugin;
import org.objectweb.fractal.bf.BindHints;
import org.objectweb.fractal.bf.BindingFactoryException;
import org.objectweb.fractal.bf.SkeletonGenerationException;
import org.objectweb.fractal.bf.proxies.StubComponentFactory;
import org.objectweb.fractal.util.Fractal;

/**
 * A plugin to import and export fractal interfaces through the Java RMI
 * protocol.
 * 
 * @since 0.3
 */
public class RmiConnector extends
		AbstractBindingFactoryPlugin<RmiExportHints, RmiBindHints> {

	Logger log = Logger.getLogger(RmiConnector.class.getCanonicalName());
	private Component bootstrap;
	private TypeFactory typeFactory;
	private GenericFactory cf;

	/**
	 * here to please Fractal..
	 */
	public RmiConnector() {
		this.initialHints = new HashMap<String, String>();
		try {
			bootstrap = Fractal.getBootstrapComponent();
			typeFactory = Fractal.getTypeFactory(bootstrap);
			cf = Fractal.getGenericFactory(bootstrap);
		} catch (Exception e) {
			throw new RmiConnectorException(e);
		}

	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#finalizeSkeleton(org.objectweb.fractal.api.Component,
	 *      java.util.Map)
	 */
	public void finalizeSkeleton(Component skel,
			Map<String, String> finalizationHints)
			throws BindingFactoryException {

	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getStubComponentFactory()
	 */
	public StubComponentFactory getStubComponentFactory() {
		return new RmiConnectorStubComponentFactoryImpl();
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#createSkel(java.lang.String,
	 *      java.lang.Object, org.objectweb.fractal.api.Component,
	 *      org.objectweb.fractal.bf.ExportHints)
	 */
	public Component createSkel(String interfaceName, Object serverItf,
			Component owner, RmiExportHints exportHints)
			throws SkeletonGenerationException {
		log.info("RmiConnector - Creating skeleton.");
		log.finest("ServerItf ref: " + serverItf + " Owner: " + owner);

		String name = exportHints.getServiceName();
		String host = exportHints.getHost();
		String port = exportHints.getPort();

		log.info("RmiConnector, export hints: \n serviceName: " + name
				+ "\n host:" + host + "\n port:" + port);

		// a fractal component whose content is a rmi skeleton.
		Component rmiServerBridge = null;

		try {

			ComponentType rmiServerBridgeType = typeFactory
					.createFcType(new InterfaceType[] {
							typeFactory
									.createFcItfType(
											"attribute-controller",
											"org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes",
											false, false, false),
							// typeFactory.createFcItfType("service",
							// "org.objectweb.fractal.api.Interface",
							// true, false, false),
							typeFactory.createFcItfType("delegate",
									"org.objectweb.fractal.api.Component",
									true, true, false) });
			rmiServerBridge = cf.newFcInstance(rmiServerBridgeType,
					"primitive",
					"org.objectweb.fractal.bf.connectors.rmi.RmiServerBridge");

			Fractal.getNameController(rmiServerBridge).setFcName(
					"org.objectweb.fractal.bf.connectors.rmi.RmiServerBridge");

			/*
			 * Set the required attributes
			 */
			RmiConnectorAttributes rmiServiceAttributes = (RmiConnectorAttributes) Fractal
					.getAttributeController(rmiServerBridge);
			rmiServiceAttributes.setServiceName(name);
			rmiServiceAttributes.setHost(host);
			rmiServiceAttributes.setPort(port);

			/*
			 * Set the binding between the rmiServerBridge and the service
			 * object
			 */
			// Interface i = (Interface) rmiServerBridge
			// .getFcInterface(RmiServerBridge.SERVICE_CLIENT_ITF_NAME);
			// BindingController bc = Fractal
			// .getBindingController(rmiServerBridge);
			// bc.bindFc(RmiServerBridge.SERVICE_CLIENT_ITF_NAME, serverItf);
			rmiServiceAttributes.setServiceObject(serverItf);
			// start the skeleton
			Fractal.getLifeCycleController(rmiServerBridge).startFc();

		} catch (Exception e) {

			throw new SkeletonGenerationException(
					"Some error occurred while exporting using the JavaRMI plugin ",
					e);
		}
		log.info("RmiConnector - skeleton created.");
		return rmiServerBridge;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getExportHints()
	 */
	public RmiExportHints getExportHints() {
		RmiExportHints rmiExportHints = new RmiExportHints();

		String registryHost = computeRegistryHost();
		rmiExportHints.setHost(registryHost);

		String registryPort = computeRegistryPort();
		rmiExportHints.setPort(registryPort);

		String serviceName = computeServiceName();
		rmiExportHints.setServiceName(serviceName);
		return rmiExportHints;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getBindHints()
	 */
	public BindHints getBindHints() {
		RmiBindHints rmiBindHints = new RmiBindHints();
		String registryHost = computeRegistryHost();
		rmiBindHints.setHost(registryHost);

		String registryPort = computeRegistryPort();
		rmiBindHints.setPort(registryPort);

		String serviceName = computeServiceName();
		rmiBindHints.setServiceName(serviceName);
		return rmiBindHints;
	}

	/**
	 * @return
	 */
	private String computeServiceName() {
		String serviceName = initialHints
				.get(RmiConnectorConstants.SERVICE_NAME);
		log.finer("Service name: " + serviceName);
		return serviceName;
	}

	/**
	 * @return
	 */
	private String computeRegistryPort() {
		String registryPort = initialHints
				.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_PORT);
		if (registryPort == null) {
			log
					.warning("You didn't specify the port for the RMI registry, the default port [1099] will be used");
			registryPort = "1099";
		}
		return registryPort;
	}

	/**
	 * @return
	 */
	private String computeRegistryHost() {
		String registryHost = initialHints
				.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS);
		if (registryHost == null) {
			log
					.warning("You didn't specify "
							+ RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS
							+ ", will use '"
							+ RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS_DEFAULT
							+ "' ");
			registryHost = RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS_DEFAULT;
		}
		return registryHost;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getPluginIdentifier()
	 */
	public String getPluginIdentifier() {
		return "rmi";
	}
}
