/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2008 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi;

import java.lang.reflect.Constructor;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.bf.AbstractSkeleton;
import org.objectweb.fractal.bf.connectors.rmi.asm.RemoteInterfaceGenerator;
import org.objectweb.fractal.bf.connectors.rmi.asm.RemoteSubclassGenerator;

/**
 * An RMI bridge to translate calls received from the underlying transport layer
 * to the business object.
 */
public class RmiServerBridge extends AbstractSkeleton implements
		RmiConnectorAttributes, BindingController, LifeCycleController {

	private String state;

	/**
	 * The name of the client interface referencing the Fractal business
	 * interface.
	 */
	public static final String SERVICE_CLIENT_ITF_NAME = "service";
	Logger log = Logger.getLogger(RmiServerBridge.class.getCanonicalName());

	/**
	 * The business object. This is a client interface, in the sense of fractal.
	 */
	private Object service;

	/**
	 * The service name to register the business object
	 */
	private String name;
	/**
	 * The host where the RMi registry is running
	 */
	private String host;
	/**
	 * The port where the RMI registry is listening to.
	 */
	private String port;

	/*
	 * Keep static references to stubs to avoid GC and DGC. XXX this might lead
	 * to memory leaks
	 */
	private static final List<Remote> server = new ArrayList<Remote>();
	/**
	 * This http server is used to let clients be able to dynamically downloaded
	 * generated stubs. There's only 1 instance of this http server shared among
	 * all the instances of the rmi server bridges. XXX A better design could be
	 * to extract this object into a component. This refactoring is delayed
	 * until it shows to be useful to someone.
	 */
	private static HttpServer HTTP_SERVER;

	private Registry registry;

	/**
	 * Default constructor.
	 */
	public RmiServerBridge() {

	}

	@SuppressWarnings("boxing")
	public void startRmi() {
		/**
		 * Only create 1 HTTP_SERVER
		 */
		if (HTTP_SERVER == null) {
			try {

				HTTP_SERVER = new HttpServer(true);
				String codebase = System
						.getProperty("java.rmi.server.codebase");
				log.info("System property: java.rmi.server.codebase ->"
						+ codebase);

			} catch (Exception e1) {
				throw new RmiConnectorException(
						"Could not start the HttpServer used for dynamic code downloading",
						e1);
			}
		}
		Remote rmiSkeleton = null;

		try {

			/*
			 * If the service reference is not Remotable
			 */
			if (!isRemotable(service)) {
				log.fine("Creating java.rmi.Remote classes for interface: "
						+ service);
				/*
				 * 1) Instantiate the transformer
				 */
				RemoteInterfaceGenerator transformer = new RemoteInterfaceGenerator();

				Class<?>[] interfaces = service.getClass().getInterfaces();
				/*
				 * 2) Use the first interface from the above array to generate a
				 * corresponding rmi
				 */
				Class<?> toTransform = interfaces[0];
				log.finest("Canonical name for java interface to transform:"
						+ toTransform);
				/*
				 * 3) Create a corresponding java class so that it extends
				 * java.rmi.Remote and that every method throws RemoteException.
				 */
				final Class<?> remotedJavaInterface = transformer
						.transform(toTransform);
				final String remotedJavaInterfaceCanonicalName = remotedJavaInterface
						.getCanonicalName();
				log.finer("RMI-enabled java interface for interface " + service
						+ " created,canonical name: "
						+ remotedJavaInterfaceCanonicalName);
				log.finest("Class loader for RMI enabled interface: "
						+ remotedJavaInterface.getClassLoader());
				/*
				 * 4) Create a generator that is capable of generating java
				 * classes that implements a given remote interface and a given
				 * non remote interface.
				 */
				RemoteSubclassGenerator remoteSubclassGenerator = new RemoteSubclassGenerator();

				Class<?> remotableClassForSkeletonContent = remoteSubclassGenerator
						.generate(toTransform, remotedJavaInterface);
				log
						.finest("Subclass implementing java.rmi.Remote and business interface: "
								+ toTransform
								+ "correctly generated: "
								+ remotableClassForSkeletonContent);
				Constructor<?> constructor;

				try {
					/*
					 * 5) found the constructor required to istantiate the
					 * content class: it's the one with one parameter whose type
					 * the same as the java interface originally to export.
					 */
					constructor = remotableClassForSkeletonContent
							.getConstructor(new Class[] { toTransform });
					log.finest("Constructor found: " + constructor);
					final Object newInstance = constructor
							.newInstance(new Object[] { service });

					log.finest("RMI Skeleton content class instantiated: "
							+ newInstance);
					// this should work here
					rmiSkeleton = (Remote) newInstance;

				} catch (Exception e) {
					e.printStackTrace();
					throw new RmiConnectorException(
							"Some error occurred while generating RMI classes for interface: "
									+ service, e);
				}
			} else {
				rmiSkeleton = (Remote) service;

			}
			RmiServerBridge.server.add(rmiSkeleton);
			Remote remote = UnicastRemoteObject.exportObject(rmiSkeleton, 0);

			if (registry == null) {
				registry = locateRegistry();
			}
			log.fine("Going to bind RMI skeleton into the RMI registry..");
			registry.bind(name, remote);
			log.fine("RMI skeleton successefully bound into registry");

		} catch (Exception e) {
			throw new RmiConnectorException(e);

		}

	}

	/**
	 * @param interfaceToExport
	 * @return
	 */
	private boolean isRemotable(Object interfaceToExport) {
		return Remote.class.isAssignableFrom(interfaceToExport.getClass());
	}

	/**
	 * @return a local reference to a remote RMI registry, if available at the
	 *         specified address and port, or a reference to a locally created
	 *         RMI registry
	 * @throws RemoteException
	 * @throw {@link RmiRegistryCreationException} if problems occur while
	 *        creating a local rmi registry.
	 */
	protected Registry locateRegistry() {
		Integer portIntValue = Integer.valueOf(port);
		if (portIntValue == null) {
			log.warning("No port set for RMI registry, will use default:"
					+ RmiConnectorConstants.JAVA_RMI_REGISTRY_DEFAULT_PORT);
			portIntValue = Integer
					.valueOf(RmiConnectorConstants.JAVA_RMI_REGISTRY_DEFAULT_PORT);
		}
		Registry registry = null;
		try {

			registry = LocateRegistry
					.getRegistry(host, portIntValue.intValue());
			log.fine("RMI registry recovered :" + registry.toString());
			/*
			 * ping the registry, to see if it's really accessible
			 */
			registry.list();

		} catch (Exception e) {
			// create it locally
			log
					.warning("Could not recover RMI registry at the "
							+ host
							+ ":"
							+ port
							+ ", now trying to create a local RMI regisrty using address localhost:"
							+ portIntValue);
			try {
				registry = LocateRegistry.createRegistry(portIntValue
						.intValue());
			} catch (RemoteException e1) {
				throw new RmiRegistryCreationException(
						"Cannot create a local RMI Registry", e1);
			}
		}

		return registry;
	}

	// attribute controller implementation
	/**
	 * @see org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes#setServiceName(java.lang.String)
	 */
	public void setServiceName(String name) {
		this.name = name;
	}

	public void setServiceObject(Object service) {
		this.service = service;
	}

	/**
	 * @see org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes#setHost(java.lang.String)
	 */
	public void setHost(String host) {
		this.host = host;
	}

	/**
	 * @see org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes#setPort(java.lang.String)
	 */
	public void setPort(String port) {
		this.port = port;
	}

	// binding-controller implementation
	@Override
	public void bindFc(String clientItfName, Object serverItf)
			throws NoSuchInterfaceException, IllegalBindingException,
			IllegalLifeCycleException {
		if (clientItfName.equalsIgnoreCase(SERVICE_CLIENT_ITF_NAME)) {
			this.service = serverItf;
		} else
			super.bindFc(clientItfName, serverItf);
	}

	@Override
	public String[] listFc() {
		String[] superItfs = super.listFc();
		String[] itfs = new String[superItfs.length + 1];
		int i = 0;
		while (i < superItfs.length) {
			itfs[i] = superItfs[i];
			i++;
		}
		itfs[i] = SERVICE_CLIENT_ITF_NAME;
		return itfs;
	}

	@Override
	public Object lookupFc(String clientItfName)
			throws NoSuchInterfaceException {
		if (clientItfName.equalsIgnoreCase(SERVICE_CLIENT_ITF_NAME)) {
			return this.service;
		} else
			return super.lookupFc(clientItfName);
	}

	@Override
	public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
			IllegalBindingException, IllegalLifeCycleException {
		if (clientItfName.equalsIgnoreCase(SERVICE_CLIENT_ITF_NAME)) {
			this.service = null;
		} else
			super.unbindFc(clientItfName);
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#getFcState()
	 */
	public String getFcState() {
		return this.state;
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
	 */
	public void startFc() throws IllegalLifeCycleException {
		startRmi();
		this.state = LifeCycleController.STARTED;
	}

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
	 */
	public void stopFc() throws IllegalLifeCycleException {
		try {
			this.registry.unbind(this.name);
			this.state = LifeCycleController.STOPPED;
		} catch (Exception e) {
			throw new RmiConnectorException(
					"Cannot unbind the service from the RMI registry", e);
		}

	}

}
