/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2008 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi.asm;

import java.io.PrintWriter;
import java.io.Serializable;
import java.io.StringWriter;
import java.lang.reflect.Method;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.logging.Logger;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.util.CheckClassAdapter;
import org.objectweb.fractal.bf.connectors.rmi.RmiConnectorException;

/**
 * A simple class transformer that given a Java interface creates a
 * corresponding Java interface, as a class object, that can be exported via
 * Java RMI. This means that:
 * <ol>
 * <li>the resulting interface extends {@link Remote},
 * <li>the resulting interface extends {@link Serializable}, if it doesn't
 * extend it already, and
 * <li>every method declared by the interface throws a {@link RemoteException}
 * </ol>
 * 
 * The idea behind this code was inspired by a similar one adopted in the
 * Binding-RMI of the Apache Tuscany project. See:
 * 
 * <pre>
 * https://svn.apache.org/repos/asf/incubator
 * /tuscany/java/sca/modules/binding-rmi
 * /src/main/java/org/apache/tuscany/sca/binding/rmi/RMIService.java
 * </pre>
 * <p>
 * Their solution adopts a mix of ASM (see http://asm.objectweb.org) and CGLIB
 * bytecode instrumentation tools. In contrast, to reduce the disk footprint for
 * those who wants to use the fractal-bf-connector-rmi plugin, this solution
 * uses ASM only.
 * </p>
 * <p>
 * The names of the java interfaces generated by this class will have as suffix:
 * {@link RemoteInterfaceGenerator#GENERATED_CLASS_SUFFIX}.
 * </p>
 * 
 * @see http://asm.objectweb.org
 * @see http://cglib.sourceforge.net/
 */
public class RemoteInterfaceGenerator {
	/**
	 * The suffix appended to the internal binary name of the generated class.
	 */
	private static final String GENERATED_CLASS_SUFFIX = "$$BindingFactoryGenerated";
	/* The logger to use for this class */
	Logger log = Logger.getLogger(RemoteInterfaceGenerator.class
			.getCanonicalName());

	/**
	 * Create a new java interface that is a loose copy of the input one, by
	 * letting it extends {@link java.rmi.Remote} and by adding a throws clause
	 * to every method so that it throws {@link RemoteException}.
	 * 
	 * @param targetJavaInterface
	 *            the java interface to be transformed
	 * @return a newly generated java interface that can be exported into the
	 *         RMI registry
	 * @throws IllegalArgumentException
	 *             if the input class is not a Java interface
	 */
	public Class<?> transform(Class<?> targetJavaInterface) {
		final String javaItfCanonicalName = targetJavaInterface
				.getCanonicalName();

		Class<?> toReturn = null;

		if (!targetJavaInterface.isInterface()) {
			throw new IllegalArgumentException("The input class : "
					+ targetJavaInterface.getCanonicalName()
					+ " is not an interface ");
		} else {

			RMIServiceClassLoader classloader = RMIServiceClassLoaderFactory
					.getRMIClassLoader();
			log.finer("JavaRMI class loader ref:  " + classloader);

			final byte[] byteCode = generateJavaRmiRemoteInterfaceBytecode(targetJavaInterface);

			StringWriter sw = new StringWriter();
			PrintWriter pw = new PrintWriter(sw);
			CheckClassAdapter.verify(new ClassReader(byteCode), false, pw);
			String cause = sw.toString();
			if (cause.length() > 0) {
				throw new RmiConnectorException(
						"An error occurred while generating an RMI stub/skeleton: \n"
								+ cause);
			}

			log.finest("Bytecode for RMI-enabled interface generated");
			log.finest("Original java interface name: " + javaItfCanonicalName);
			final String generatedInterfaceInternalName = javaItfCanonicalName
					+ GENERATED_CLASS_SUFFIX;
			log.finest("Generated java interface name: "
					+ generatedInterfaceInternalName);

			Class<?> remotedTargetJavaInterface = classloader.defineClass(
					byteCode, generatedInterfaceInternalName);
			log.finest("Class " + remotedTargetJavaInterface.getCanonicalName()
					+ " for RMI-enabled interface defined in the "
					+ RMIServiceClassLoader.class.getName());
			toReturn = remotedTargetJavaInterface;
		}

		return toReturn;
	}

	/**
	 * Generate the bytecode. Defining the class from the byte code must be the
	 * responsibility of the caller of this method, since it requires a
	 * ClassLoader to be created to define and load this interface.
	 * 
	 * @param serviceInterface
	 *            the input interface whose corresponding Java RMI enabled
	 *            interface has to be generated.
	 */
	protected byte[] generateJavaRmiRemoteInterfaceBytecode(
			Class<?> serviceInterface) {
		String interfazeName = serviceInterface.getCanonicalName();
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);

		final String internalNameGeneratedClass = interfazeName.replace('.',
				'/')
				+ GENERATED_CLASS_SUFFIX;
		log.finest("Internal name for generated class: "
				+ internalNameGeneratedClass);
		String[] implementedInterfaces = computeInterfacesfrom(serviceInterface);
		cw.visit(Opcodes.V1_5, Opcodes.ACC_PUBLIC + Opcodes.ACC_ABSTRACT
				+ Opcodes.ACC_INTERFACE, internalNameGeneratedClass, null,
				"java/lang/Object", implementedInterfaces);

		StringBuffer argsAndReturn = null;
		Method[] methods = serviceInterface.getMethods();
		for (Method method : methods) {
			argsAndReturn = new StringBuffer("(");
			Class<?>[] paramTypes = method.getParameterTypes();
			Class<?> returnType = method.getReturnType();

			for (Class<?> paramType : paramTypes) {
				argsAndReturn.append(Type.getType(paramType));
			}
			argsAndReturn.append(")");
			argsAndReturn.append(Type.getType(returnType));

			cw.visitMethod(Opcodes.ACC_PUBLIC + Opcodes.ACC_ABSTRACT, method
					.getName(), argsAndReturn.toString(), null,
					new String[] { "java/rmi/RemoteException" });
		}
		cw.visitEnd();
		return cw.toByteArray();
	}

	/**
	 * Return an array of strings containing the binary names of the java
	 * interfaces to be implemented by the generated class. The element at index
	 * 0 is always "java/rmi/Remote". The element at index 1 is always the
	 * internal name of the serviceInterface parameter. If the input class
	 * doesn't implement {@link Serializable}, the element at index 2 is
	 * "java/io/Serializable".
	 * 
	 * @param serviceInterface
	 * @return an array of strings containing the binary names of the java
	 *         interfaces to be implemented by the generated class
	 */
	private String[] computeInterfacesfrom(Class<?> serviceInterface) {
		String[] itfs;
		final boolean isNotSerializable = !Serializable.class
				.isAssignableFrom(serviceInterface);
		if (isNotSerializable) {
			itfs = new String[3];
		} else {
			itfs = new String[2];
		}
		itfs[0] = "java/rmi/Remote";
		itfs[1] = Type.getInternalName(serviceInterface);
		if (isNotSerializable) {
			itfs[2] = "java/io/Serializable";
		}
		return itfs;
	}
}
