/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2008 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi.asm;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Method;
import java.util.logging.Logger;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.util.CheckClassAdapter;
import org.objectweb.fractal.bf.connectors.rmi.RmiConnectorException;

/**
 * Generate a class that implements the generated subclass of
 * {@link java.rmi.Remote}.
 */
public class RemoteSubclassGenerator implements Opcodes {

	/**
	 * The name of the field to generate referencing the object to which
	 * delegate calls.
	 */
	private static final String DELEGATED_FIELD_NAME = "s";
	Logger log = Logger.getLogger(RemoteSubclassGenerator.class
			.getCanonicalName());

	/**
	 * Generate and instantiate a class that can be exported into a RMI
	 * registry.
	 * 
	 * @param unremotableInterfaze
	 *            the type of the delegate field to which calls will be
	 *            delegated to
	 * @param remotableInterfaze
	 *            a subclass of java.rmi.Remote, that is implemented by the
	 *            resulting class
	 * @throws RmiConnectorException
	 *             if the generated bytecode doesn't pass the checks performed
	 *             by
	 *             {@link CheckClassAdapter#verify(ClassReader, boolean, PrintWriter)}
	 *             method.
	 * @see CheckClassAdapter
	 */
	public Class<?> generate(Class<?> unremotableInterfaze,
			Class<?> remotableInterfaze) {
		final String unremotableInterfazeInternalName = Type
				.getInternalName(unremotableInterfaze);
		final String remotableInterfazeInternalName = Type
				.getInternalName(remotableInterfaze);

		log.finest("Internal type of delegate field: "
				+ unremotableInterfazeInternalName);
		/*
		 * Using COMPUTE_MAXS flag avoids to handle manually stack indices, to
		 * be specified in the various computeMaxs methods.
		 */
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);

		FieldVisitor fv;
		MethodVisitor mv;

		final String generatedClassInternalName = remotableInterfazeInternalName
				+ "Impl";
		log.finest("Generated class internal name: "
				+ generatedClassInternalName);
		cw.visit(V1_5, ACC_PUBLIC + ACC_SUPER, generatedClassInternalName,
				null, "java/lang/Object",
				new String[] { remotableInterfazeInternalName,
						"java/io/Serializable" });

		{

			fv = cw.visitField(ACC_PUBLIC, DELEGATED_FIELD_NAME, "L"
					+ unremotableInterfazeInternalName + ";", null, null);
			fv.visitEnd();
			log.finest("Field name '" + DELEGATED_FIELD_NAME + "' of type "
					+ unremotableInterfazeInternalName + " generated");
		}
		/*
		 * Generate 1-arg constructor. The type of the parameter is the same as
		 * the unremotableInterfaze parameter.
		 */
		{
			mv = cw.visitMethod(ACC_PUBLIC, "<init>", "(L"
					+ unremotableInterfazeInternalName + ";)V", null,
					new String[] { "java/rmi/RemoteException" });
			mv.visitCode();
			Label l0 = new Label();
			mv.visitLabel(l0);
			mv.visitVarInsn(ALOAD, 0);
			mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>",
					"()V");
			Label l1 = new Label();
			mv.visitLabel(l1);
			mv.visitVarInsn(ALOAD, 0);
			mv.visitVarInsn(ALOAD, 1);
			mv.visitFieldInsn(PUTFIELD, generatedClassInternalName,
					DELEGATED_FIELD_NAME, "L"
							+ unremotableInterfazeInternalName + ";");
			Label l2 = new Label();
			mv.visitLabel(l2);
			mv.visitInsn(RETURN);
			Label l3 = new Label();
			mv.visitLabel(l3);
			mv.visitLocalVariable("this", "L" + generatedClassInternalName
					+ ";", null, l0, l3, 0);
			mv.visitLocalVariable(DELEGATED_FIELD_NAME, "L"
					+ unremotableInterfazeInternalName + ";", null, l0, l3, 1);
			mv.visitMaxs(2, 2);
			mv.visitEnd();
			log.fine("1-Args constructor generated.");

		}
		/*
		 * for every method declared by the remotableInterfaze interface,
		 * generate a method in the class being generated with the same
		 * signature
		 */
		for (Method method : remotableInterfaze.getMethods()) {
			log.finest("Generating method body for " + method.getName());

			Class<?>[] paramTypes = method.getParameterTypes();
			int args = paramTypes.length;

			Class<?> returnType = method.getReturnType();
			String desc = Type.getMethodDescriptor(method);
			mv = cw.visitMethod(ACC_PUBLIC, method.getName(), desc, null,
					new String[] { "java/rmi/RemoteException" });

			mv.visitCode();

			Label l0 = new Label();
			mv.visitLabel(l0);
			/*
			 * push this on the stack
			 */
			mv.visitVarInsn(ALOAD, 0);
			/*
			 * recover a reference to the delegate field for 'this' object
			 */
			mv.visitFieldInsn(GETFIELD, generatedClassInternalName,
					DELEGATED_FIELD_NAME, "L"
							+ unremotableInterfazeInternalName + ";");
			Label l1 = new Label();
			mv.visitLabel(l1);
			mv.visitLocalVariable("this", "L" + generatedClassInternalName
					+ ";", null, l0, l1, 0);
			/*
			 * for every parameter of this method, push the actual arguments on
			 * the stack. Depending on the type of the variable, a corresponding
			 * java-type dependant LOAD instruction must be executed.
			 */
			log.finest("There are " + args + " arguments to push on stack");
			int i = 1;
			for (Class<?> c : paramTypes) {
				Type paramType = Type.getType(c);
				int opcode = paramType.getOpcode(ILOAD);
				final int size = paramType.getSize();
				log.finest("Next index used by visitVarIns: " + i);
				/* declares local variables contextually to the pushing */
				/*
				 * The name is not important as its used only for debug infos,
				 * Also, there's no way to recover the original name of the
				 * variable as it was used in the original code no support in
				 * the reflection APIs. A deeper discussion might be found
				 * following links here:
				 * http://www.jroller.com/vschiavoni/entry/how_to_get_the_name
				 */
				mv.visitLocalVariable("a", paramType.toString(), null, l0, l1,
						i);
				mv.visitVarInsn(opcode, i);
				i = i + size;
			}

			mv.visitMethodInsn(INVOKEINTERFACE,
					unremotableInterfazeInternalName, method.getName(), desc);

			final String returnTypeValue = returnType.toString();

			if (returnTypeValue.equalsIgnoreCase("void")) {
				log.finest("Method " + method
						+ " returns void, using RETURN opcode");
				// XXX should use the proper Opcode returned by Type.getOpcode
				mv.visitInsn(RETURN);
			} else {
				int opcodeReturn = Type.getType(returnType).getOpcode(IRETURN);
				log.finest("Method " + method + " return " + returnTypeValue
						+ ", using ARETURN opcode");
				mv.visitInsn(opcodeReturn);
			}

			/* this should not be required because of the COMPUTE_MAX */
			mv.visitMaxs(0, 0);
			mv.visitEnd();

			log.finer("Method " + method.toGenericString() + " generated");
		}

		cw.visitEnd();
		byte[] bytecode = cw.toByteArray();
		/*
		 * check that the bytecode is valid.
		 */
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		CheckClassAdapter.verify(new ClassReader(bytecode), false, pw);
		String cause = sw.toString();
		if (cause.length() > 0) {
			throw new RmiConnectorException(
					"An error occurred while generating an RMI stub/skeleton: \n"
							+ cause);
		}

		RMIServiceClassLoader classloader = RMIServiceClassLoaderFactory
				.getRMIClassLoader();
		log.fine("RMIServiceClassLoader found: " + classloader);
		log.finest("RMIServiceClassLoader, classes currently defined: "
				+ classloader.getDefinedClasses());
		final Class<?> definedClazz = classloader.defineClass(bytecode,
				generatedClassInternalName.replace("/", "."));
		log.finest("RMIServiceClassLoader just defined class: " + definedClazz);

		return definedClazz;

	}
}
